from .spells import Spell


class WallOfFire(Spell):
    """You create a wall of fire on a solid surface within range. You can make the wall
     up to 60 feet long, 20 feet high, and 1 foot thick, or a ringed wall up to 20 
    feet in diameter, 20 feet high, and 1 foot thick. The wall is opaque and lasts 
    for the duration.
    
    When the wall appears, each creature within its area must 
    make a Dexterity saving throw. On a failed save, a creature takes 5d8 fire 
    damage, or half as much damage on a successful save.
    
    One side of the wall, 
    selected by you when you cast this spell, deals 5d8 fire damage to each creature
     that ends its turn within 10 feet of that side or inside the wall. A creature 
    takes the same damage when it enters the wall for the first time on a turn or 
    ends its turn there. The other side of the wall deals no damage.
    
    At Higher 
    Levels: When you cast this spell using a spell slot of 5th level or higher, the 
    damage increases by 1d8 for each slot level above 4th.
    """
    name = "Wall Of Fire"
    level = 4
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ('V', 'S', 'M')
    materials = """A small piece of phosphorus"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Evocation"
    classes = ('Druid', 'Sorcerer', 'Wizard')


class WallOfForce(Spell):
    """An invisible wall of force springs into existence at a point you choose within 
    range.
    The wall appears in any orientation you choose, as a horizontal or 
    vertical barrier or at an angle. It can be free floating or resting on a solid 
    surface. You can form it into a hemispherical dome or a sphere with a radius of 
    up to 10 feet, or you can shape a flat surface made up of ten 10-foot-by-10-foot
     panels. Each panel must be continguous with another panel. In any form, the 
    wall is 1/4 inch thick. It lasts for the duration. If the wall cuts through a 
    creature’s space when it appears, the creature is pushed to one side of the wall
     (your choice which side).
    
    Nothing can physically pass through the wall. It is 
    immune to all damage and can’t be dispelled by dispel magic. A disintegrate 
    spell destroys the wall instantly, however.  The wall also extends into the 
    Ethereal Plane, blocking ethereal travel through the wall.
    """
    name = "Wall Of Force"
    level = 5
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ('V', 'S', 'M')
    materials = """A pinch of powder made by crushing a clear gemstone"""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Evocation"
    classes = ('Wizard',)


class WallOfIce(Spell):
    """You create a wall of ice on a solid surface within range. You can form it into a
     hemispherical dome or a sphere with radius of up to 10 feet, or you can shape a
     flat surfcae made up of ten 10-foot-square panels. Each panel must be 
    contiguous with another panel. In any form, the wall is 1 foot thick and lasts 
    for the duration.
    
    If the wall cuts through a creature’s space when it appears, 
    the creature within its area is pushed to one side of the wall and must make a 
    Dexterity saving throw. On a failed save, the creature takes 10d6 cold damage, 
    or half as much damage on a successful save.
    
    The wall is an object that can be 
    damaged and thus breached. It has AC 12 and 30 hit points per 10-foot section, 
    and it is vulnerable to fire damage. Reducing a 10-foot section of wall to 0 hit
     points destroys it and leaves behind a sheet of frigid air int he space the 
    wall occupied. A creature moving through the sheet of frigid air for the first 
    time on a turn must make a Constitution saaving throw. The creature takes 5f6 
    cold damage on a failed save, or half as much damage on a successful one.
    
    At 
    Higher Levels: When you cast this spell using a spell slot of 7th level or 
    higher, the damage the wall deals when it appears increases by 2d6, and the 
    damage from passing through the sheet of frigid air increases by 1d6, for each 
    slot level above 6th.
    """
    name = "Wall Of Ice"
    level = 6
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ('V', 'S', 'M')
    materials = """A small piece of quartz"""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Evocation"
    classes = ('Wizard',)


class WallOfLight(Spell):
    """A shimmering wall of bright light appears at a point you choose within range. 
    The wall appears in any orientation you choose: horizontally, vertically, or 
    diagonally. It can be free floating, or it can rest on a solid surface. The wall
     can be up to 60 feet long, 10 feet high, and 5 feet thick. The wall blocks line
     of sight, but creatures and objects can pass through it. It emits bright light 
    out to 120 feet and dim light for an additional 120 feet.
    When the wall appears,
     each creature in its area must make a Constitution saving throw. On a failed 
    save, a creature takes 4d8 radiant damage, and it is blinded for 1 minute. On a 
    successful save, it takes half as much damage and isn’t blinded. A blinded 
    creature can make a Constitution saving throw at the end of each of its turns, 
    ending the effect on itself on a success.
    A creature that ends its turn in the 
    wall’s area takes 4d8 radiant damage.
    Until the spell ends, you can use an 
    action to launch a beam of radiance from the wall at one creature you can see 
    within 60 feet of it. Make a ranged spell attack. On a hit, the target takes 4d8
     radiant damage. Whether you hit or miss, reduce the length of the wall by 10 
    feet. If the wall’s length drops to 0 feet, the spell ends.
    
    At Higher Levels: 
    When you cast this spell using a spell slot of 6th level or higher, the damage 
    increases by 1d8 for each slot level above 5th.
    """
    name = "Wall Of Light"
    level = 5
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ('V', 'S', 'M')
    materials = """A hand mirror"""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Evocation"
    classes = ('Sorcerer', 'Warlock', 'Wizard')


class WallOfSand(Spell):
    """You conjure up a wall of swirling sand on the ground at a point you can see 
    within range. You can make the wall up to 30 feet long, 10 feet high, and 10 
    feet thick, and it vanishes when the spell ends. It blocks line of sight but not
     movement. A creature is blinded while in the wall’s space and must spend 3 feet
     of movement for every 1 foot it moves there.
    """
    name = "Wall Of Sand"
    level = 3
    casting_time = "1 action"
    casting_range = "90 feet"
    components = ('V', 'S', 'M')
    materials = """A handful of sand"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ('Wizard',)


class WallOfStone(Spell):
    """A nonmagical wall of solid stone springs into existence at a point you choose 
    within range.
    The wall is 6 inches thick and is composed of ten 10-foot-
    by-10-foot panels. Each panel must be contiguous with at least on other panel. 
    Alternatively, you can create 10-foot-by-20-foot panels that are only 3 inches 
    thick.
    
    If the wall cuts through a creature’s space when it appears, the 
    creature is pushed to one side of the wall (your choice). If a creature would be
     surrounded on all sides by the wall (or the wall and another solid surface), 
    that creature can make a Dexterity saving throw. On a success, it can use its 
    reaction to move up to its speed so that it is no longer enclosed by the wall.
    
    
    The wall can have any shape you desire, though it can’t occupy the same space as
     a creature or object. the wall doesn’t need to be vertical or resting on any 
    firm foundation. It must, however, merge with and be solidly supported by 
    existing stone. Thus you can use this spell to bridge a chasm or create a ramp.
    
    
    If you create a span greater than 20 feet in length, you must halve the size 
    of each panel to create supports. You can crudely shape the wall to create 
    crenellations, battlements, and so on.
    
    The wall is an object made of stone that
     can be damaged and thus breached. Each panel has AC 15 and 30 hit points per 
    inch of thickness. Reducing a panel to 0 hit points destroys it and might cause 
    connected panels to collapse at the DM’s discretion.
    
    If you maintain your 
    concentration on this spell for its whole duration, the wall becomes permanent 
    and can’t be dispelled. Otherwise, the wall disappears when the spell ends.
    """
    name = "Wall Of Stone"
    level = 5
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ('V', 'S', 'M')
    materials = """A small block of granite"""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Evocation"
    classes = ('Druid', 'Sorcerer', 'Wizard')


class WallOfThorns(Spell):
    """You create a wall of tough, pliable, tangled brush bristling with needle-sharp 
    thorns. The wall appears within range on a solid surface and lasts for the 
    duration. You choose to make the wall up to 60 feet long, 10 feet high, and 5 
    feet thick or a circle that has a 20-foot diameter and is up to 20 feet high and
     5  feet thick. The wall blocks line of sight.
    
    When the wall appears, each 
    creature within its area must make a Dexterity saving throw. On a failed save, a
     creature takes 7d8 piercing damage, or half as much damage on a successful 
    save.
    
    A creature can move through the wall, albeit slowly and painfully. For 
    every 1 foot a creature moves through the wall, it must spend 4 feet of 
    movement. Furthermore, the first time a creature enters the wall on a turn or 
    ends its turn there, the creature must make a Dexterity saving throw. It takes 
    7d8 slashing damage on a failed save, or half as much on a successful save.
    
    At 
    Higher Levels: When you cast this spell using a spell slot of 7th level or 
    higher, both types o f damage increase by 1d8 for each slot level above 6th.
    """
    name = "Wall Of Thorns"
    level = 6
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ('V', 'S', 'M')
    materials = """A handful of thorns"""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Druid',)


class WallOfWater(Spell):
    """(a drop of water)
    You conjure up a wall of water on the ground at a point you 
    can see within range. You can make the wall up to 30 feet long, 10 feet high, 
    and 1 foot thick, or you can make a ringed wall up to 20 feet in diameter, 20 
    feet high, and 1 foot thick. The wall vanishes when the spell ends. The wall’s 
    space is difficult terrain.
    Any ranged weapon attack that enters the wall’s 
    space has disadvantage on the attack roll, and fire damage
    is halved if the fire
     effect passes through the wall to reach its target. Spells that deal cold 
    damage that pass through the wall cause the area of the wall they pass through 
    to freeze solid (at least a 5-foot square section is frozen). Each 5-foot-square
     frozen section has AC 5 and 15 hit points. Reducing a frozen section to 0 hit 
    points destroys it. When a section is destroyed, the wall’s water doesn’t fill 
    it.
    """
    name = "Wall Of Water"
    level = 3
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ('V', 'S', 'M')
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ('Druid', 'Sorcerer', 'Wizard')


class WardingBond(Spell):
    """This spell wards a willing creature you touch and creates a mystic connection 
    between you and the target until the spell ends.
    
    While the target is within 60 
    feet of you, it gains a +1 bonus to AC and saving throws, and it has resistance 
    to all damage. Also, each time it takes damage, you take the same amount of 
    damage.
    
    The spell ends if you drop to 0 hit points or if you and the target 
    become separated by more than 60 feet. It also ends if the spell is cast again 
    on either of the connected creatures. You can also dismiss the spell as an 
    action.
    """
    name = "Warding Bond"
    level = 2
    casting_time = "1 action"
    casting_range = "Touch"
    components = ('V', 'S', 'M')
    materials = """A pair of platinum rings worth at least 50 gp each, which you and target must wear for the duration"""
    duration = "1 hour"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Cleric',)


class WardingWind(Spell):
    """A strong wind (20 miles per hour) blows around you in a 10-foot radius and moves
     with you, remaining centered on you. The wind lasts for the spell’s duration.
    
    The wind has the following effects:
    • It deafens you and other creatures in its 
    area.
    • It extinguishes unprotected flames in its area that are torch-sized or 
    smaller.
    • The area is difficult terrain for creatures other than you.
    • The 
    attack rolls of ranged weapon attacks have disadvantage if they pass in or out 
    of the wind.
    • It hedges out vapor, gas, and fog that can be dispersed by strong
     wind.
    """
    name = "Warding Wind"
    level = 2
    casting_time = "1 action"
    casting_range = "Self"
    components = ('V',)
    materials = """"""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Evocation"
    classes = ('Bard', 'Druid', 'Sorcerer', 'Wizard')


class WaterBreathing(Spell):
    """This spell grants up to ten willing creatures you can see within range the 
    ability to breathe underwater until the spell ends. Affected creatures also 
    retain their normal mode of respiration.
    """
    name = "Water Breathing"
    level = 3
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ('V', 'S', 'M')
    materials = """A short reed or piece of straw"""
    duration = "24 hours"
    ritual = True
    magic_school = "Transmutation"
    classes = ('Druid', 'Ranger', 'Sorcerer', 'Wizard')


class WaterWalk(Spell):
    """This spell grants the ability to move across any liquid surface – such as water,
     acid, mud, snow, quicksand, or lava – as if it were harmless solid ground 
    (creatures crossing molten lava can still take damage from the heat).
    Up to ten 
    willing creatures you can see within range gain this ability for the duration.
    
    
    If you target a creature submerged in a liquid, the spell carries the target to 
    the surface of the liquid at a rate of 60 feet per round.
    """
    name = "Water Walk"
    level = 3
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ('V', 'S', 'M')
    materials = """A piece of cork"""
    duration = "1 hour"
    ritual = True
    magic_school = "Transmutation"
    classes = ('Cleric', 'Druid', 'Ranger', 'Sorcerer')


class WaterySphere(Spell):
    """You conjure up a sphere of water with a 5-foot radius on a point you can see 
    within range. The sphere can hover in the air, but no more than 10 feet off the 
    ground. The sphere remains for the spell’s duration.
    Any creature in the 
    sphere’s space must make a Strength saving throw. On a successful save, a 
    creature is ejected from that space to the nearest unoccupied space outside it. 
    A Huge or larger creature succeeds on the saving throw automatically. On a 
    failed save, a creature is restrained by the sphere and is engulfed by the 
    water. At the end of each of its turns, a restrained target can repeat the 
    saving throw.
    The sphere can restrain a maximum of four Medium or smaller 
    creatures or one Large creature. If the sphere restrains a creature in excess of
     these numbers, a random creature that was already restrained by the sphere 
    falls out of it and lands prone in a space within 5 feet of it.
    As an action, 
    you can move the sphere up to 30 feet in a straight line. If it moves over a 
    pit, cliff, or other drop, it safely descends until it is hovering 10 feet over 
    ground. Any creature restrained by the sphere moves with it. You can ram the 
    sphere into creatures, forcing them to make the saving throw, but no more than 
    once per turn.
    When the spell ends, the sphere falls to the ground and 
    extinguishes all normal flames within 30 feet of it. Any creature restrained by 
    the sphere is knocked prone in the space where it falls.
    """
    name = "Watery Sphere"
    level = 4
    casting_time = "1 action"
    casting_range = "90 feet"
    components = ('V', 'S', 'M')
    materials = """A droplet of water"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Druid', 'Sorcerer', 'Wizard')


class Web(Spell):
    """You conjure a mass of thick, sticky webbing at a point of your choice within 
    range.
    The webs fill a 20-foot cube from that point for the duration. The webs 
    are difficult terrain and lightly obscure their area.
    
    If the webs aren’t 
    anchored between two solid masses (such as walls or trees) or layered across a 
    floor, wall, or ceiling, the conjured web collapses on itself, and the spell 
    ends at the start of your next turn. Webs layered over a flat surface have a 
    depth of 5 feet.
    
    Each creature that starts its turn in the webs or that enters 
    them during its turn must make a Dexterity saving throw. On a failed save, the 
    creature is restrained as long as it remains in the webs or until it breaks 
    free.
    
    A creature restrained by the webs can use its action to make a Strength 
    check against your spell save DC. If it succeeds, it is no longer restrained.
    
    
    The webs are flammable. Any 5-foot cube of webs exposed to fire burns away in 1 
    round, dealing 2d4 fire damage to any creature that starts its turn in the fire.
    """
    name = "Web"
    level = 2
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ('V', 'S', 'M')
    materials = """A bit of spiderweb"""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Sorcerer', 'Wizard')


class Weird(Spell):
    """Drawing on the deepest fears of a group of creatures, you create illusory 
    creatures in their minds, visible only to them.
    Each creature in a 30-foot-
    radius sphere centered on a point of your choice within range must make a Wisdom
     saving throw. On a failed save, a creature becomes frightened for the duration.
    
    The illusion calls on the creature’s deepest fears, manifesting its worst 
    nightmares as an implacable threat. At the end of each of the frightened 
    creature’s turns, it must succeed on a Wisdom saving throw or take 4d10 psychic 
    damage. On a successful save, the spell ends for that creature.
    """
    name = "Weird"
    level = 9
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Illusion"
    classes = ('Wizard',)


class Whirlwind(Spell):
    """A whirlwind howls down to a point that you can see on the ground within range. 
    The whirlwind is a 10-foot-radius, 30-foot-high cylinder centered on that point.
     Until the spell ends, you can use your action to move the whirlwind up to 30 
    feet in any direction along the ground. The whirlwind sucks up any Medium or 
    smaller objects that aren’t secured to anything and that aren’t worn or carried 
    by anyone.
    A creature must make a Dexterity saving throw the first time on a 
    turn that it enters the
    whirlwind or that the whirlwind enters its space, 
    including when the whirlwind first appears. A creature takes 10d6 bludgeoning 
    damage on a failed save, or half as much damage on a successful one. In 
    addition, a Large or smaller creature that fails the save must succeed on a 
    Strength saving throw or become restrained in the whirlwind until the spell 
    ends. When a creature starts its turn restrained by the whirlwind, the creature 
    is pulled 5 feet higher inside it, unless the creature is at the top.
    A 
    restrained creature moves with the whirlwind and falls when the spell ends, 
    unless the creature has some means to stay aloft. A restrained creature can use 
    an action to make a Strength or Dexterity check against your spell save DC. If 
    successful, the creature is no longer restrained by the whirlwind and is hurled 
    3d6 × 10 feet away from it in a random direction.
    """
    name = "Whirlwind"
    level = 7
    casting_time = "1 action"
    casting_range = "300 feet"
    components = ('V', 'M')
    materials = """A piece of straw"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Evocation"
    classes = ('Druid', 'Wizard', 'Sorcerer')


class WindWalk(Spell):
    """You and up to ten willing creatures you can see within range assume a gaseous 
    form for the duration, appearing as wisps of cloud.
    While in this cloud form, a 
    creature has a flying speed of 300 feet and has resistance to damage from 
    nonmagical weapons. The only actions a creature can take in this form are the 
    Dash action or to revert to its normal form.
    Reverting takes 1 minute, during 
    which time a creature is incapacitated and can’t move. Until the spell ends, a 
    creature can revert to cloud form, which also requires the 1-minute 
    transformation.
    
    If a creature is in cloud form and flying when the effect ends,
     the creature descends 60 feet per round for 1 minute until it lands, which it 
    does safely. If it can’t land after 1 minute, the creature falls the remaining 
    distance.
    """
    name = "Wind Walk"
    level = 6
    casting_time = "1 minute"
    casting_range = "30 feet"
    components = ('V', 'S', 'M')
    materials = """Fire and holy water"""
    duration = "8 hours"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Druid',)


class WindWall(Spell):
    """A wall of strong wind rises from the ground at a point you choose within range.
    
    You can make the wall up to 50 feet long, 15 feet high, and 1 foot thick. You 
    can shape the wall in any way you choose so long as it makes one continuous path
     along the ground. The wall lasts for the duration.
    
    When the wall appears, each
     creature within its area must make a Strength saving throw. A creature takes 
    3d8 bludgeoning damage on a failed save, or half as much damage on a successful 
    one.
    
    The strong wind keeps fog, smoke, and other gases at bay. Small or smaller
     flying creatures or objects can’t pass through the wall. Loose, lightweight 
    materials brought into the wall fly upward. Arrows, bolts, and other ordinary 
    projectiles launched at targets behind the wall are deflected upward and 
    automatically miss. (Boulders hurled by giants or siege engines, and similar 
    projectiles, are unaffected.) Creatures in gaseous form can’t pass through it.
    """
    name = "Wind Wall"
    level = 3
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ('V', 'S', 'M')
    materials = """A tiny fan and a feather of exotic origin"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Evocation"
    classes = ('Druid', 'Ranger')


class Wish(Spell):
    """Wish is the mightiest spell a mortal creature can cast. By simply speaking 
    aloud, you can alter the very foundations of reality in accord with your 
    desires.
    
    The basic use of this spell is to duplicate any other spell of 8th 
    level or lower. You don’t need to meet any requirements in that spell, including
     costly components. The spell simply takes effect.
    Alternatively, you can create
     one of the following effects of your choice:
    
    • You create one object of up to 
    25,000 gp in value that isn’t a magic item. The object can be no more than 300 
    feet in any dimension, and it appears in an unoccupied space you can see on the 
    ground.
    
    • You allow up to twenty creatures that you can see to regain all hit 
    points, and you end all effects on them described in the greater restoration 
    spell.
    
    • You grant up to ten creatures that you can see resistance to a damage 
    type you choose.
    
    • You grant up to ten creatures you can see immunity to a 
    single spell or other magical effect for 8 hours. For instance, you could make 
    yourself and all your com panions immune to a lich’s life drain attack.
    
    • You 
    undo a single recent event by forcing a reroll of any roll made within the last 
    round (including your last turn). Reality reshapes itself to accommodate the new
     result. For example, a wish spell could undo an opponent’s successful save, a 
    foe’s critical hit, or a friend’s failed save. You can force the reroll to be 
    made with advantage or disadvantage, and you can choose whether to use the 
    reroll or the original roll.
    
    You might be able to achieve something beyond the 
    scope of the above examples. State your wish to the DM as precisely as possible.
     The DM has great latitude in ruling what occurs in such an instance; the 
    greater the wish, the greater the likelihood that something goes wrong. This 
    spell might simply fail, the effect you desire 
    mightonlybepartlyachieved,oryoumightsuffersome unforeseen consequence as a 
    result of how you worded the wish. For example, wishing that a villain were dead
     might propel you forward in time to a period when that villain is no longer 
    alive, effectively removing you from the game. Similarly, wishing for a 
    legendary magic item or artifact might instantly transport you to the presence 
    of the item’s current owner.
    
    The stress of casting this spell to produce any 
    effect other than duplicating another spell weakens you. After enduring that 
    stress, each time you cast a spell until you finish a long rest, you take 1d10 
    necrotic damage per level of that spell. This damage can’t be reduced or 
    prevented in any way. In addition, your Strength drops to 3, if it isn’t 3 or 
    lower already, for 2d4 days. For each of those days that you spend resting and 
    doing nothing more than light activity, your remaining recovery time decreases 
    by 2 days. Finally, there is a 33 percent chance that you are unable to cast 
    wish ever again if you suffer this stress.
    """
    name = "Wish"
    level = 9
    casting_time = "1 action"
    casting_range = "Self"
    components = ('V',)
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Sorcerer', 'Wizard')


class WitchBolt(Spell):
    """A beam of crackling, blue energy lances out toward a creature within range, 
    forming a sustained arc of lightning between you and the target.
    Make a ranged 
    spell attack against that creature. On a hit, the target takes 1d12 lightning 
    damage, and on each of your turns for the duration, you can use your action to 
    deal 1d12 lightning damage to the target automatically. The spell ends if you 
    use your action to do anything else. The spell also ends if the target is ever 
    outside the spell’s range or if it has total cover from you.
    
    At Higher Levels: 
    When you cast this spell using a spell slot of 2nd level or higher, the initial 
    damage increases by 1d12 for each slot level above 1st.
    """
    name = "Witch Bolt"
    level = 1
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ('V', 'S', 'M')
    materials = """A twig from a tree that has been struck by lightning"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Evocation"
    classes = ('Sorcerer', 'Warlock', 'Wizard')


class WordOfRadiance(Spell):
    """You utter a divine word, and burning radiance erupts from you. Each creature of 
    your choice that you can see within range must succeed on a Constitution saving 
    throw or take 1d6 radiant damage.
    The spell’s damage increases by 1d6 when you 
    reach 5th level (2d6), 11th level (3d6), and 17th level (4d6).
    """
    name = "Word Of Radiance"
    level = 0
    casting_time = "1 action"
    casting_range = "5 feet"
    components = ('V', 'M')
    materials = """A holy symbol"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ('Cleric',)


class WordOfRecall(Spell):
    """You and up to five willing creatures within 5 feet of you instantly teleport to 
    a previously designated sanctuary.
    You and any creatures that teleport with you 
    appear in the nearest unoccupied space to the spot you designated when you 
    prepared your sanctuary (see below). If you cast this spell without first 
    preparing a sanctuary, the spell has no effect.
    
    You must designate a sanctuary 
    by casting this spell within a location, such as a temple, dedicated to or 
    strongly linked to your deity. If you attempt to cast the spell in this manner 
    in an area that isn’t dedicated to your deity, the spell has no effect.
    """
    name = "Word Of Recall"
    level = 6
    casting_time = "1 action"
    casting_range = "5 feet"
    components = ('V',)
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Cleric',)


class WrathOfNature(Spell):
    """You call out to the spirits of nature to rouse them against your enemies. Choose
     a point you can see within range. The spirits cause trees, rocks, and grasses 
    in a 60-foot cube centered on that point to become animated until the spell 
    ends.
    Grasses and Undergrowth. Any area of ground in the cube that is covered by
     grass or undergrowth is difficult terrain for your enemies.
    Trees. At the start
     of each of your turns, each of your enemies within 10 feet of any tree in the 
    cube must succeed on a Dexterity saving throw or take 4d6 slashing damage from 
    whipping branches.
    Roots and Vines. At the end of each of your turns, one 
    creature of your choice that is on the ground in the cube must succeed on a 
    Strength saving throw or become restrained until the spell ends. A restrained 
    creature can use an action to make a Strength (Athletics) check against your 
    spell save DC, ending the effect on itself on a success.
    Rocks. As a bonus 
    action on your turn, you can cause a loose rock in the cube to launch at a 
    creature you can see in the cube. Make a ranged spell attack against the target.
     On a hit, the target takes 3d8 nonmagical bludgeoning damage, and it must 
    succeed on a Strength saving throw or fall prone.
    """
    name = "Wrath Of Nature"
    level = 5
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Evocation"
    classes = ('Druid', 'Ranger')


class WrathfulSmite(Spell):
    """The next time you hit with a melee weapon attack during this spell’s duration, 
    your attack deals an extra 1d6 psychic damage.
    Additionally, if the target is a 
    creature, it must make a Wisdom saving throw or be frightened of you until the 
    spell ends. As an action, the creature can make a Wisdom check against your 
    spell save DC to steel its resolve and end this spell.
    """
    name = "Wrathful Smite"
    level = 1
    casting_time = "1 bonus action"
    casting_range = "Self"
    components = ('V',)
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Evocation"
    classes = ('Paladin',)





#!/usr/bin/env python

import logging
log = logging.getLogger(__name__)
import argparse
import importlib.util
import os
import subprocess
import warnings
import re
from io import StringIO

from fdfgen import forge_fdf
import pdfrw
from jinja2 import Environment, PackageLoader

from . import character as _char
from . import exceptions, classes
from .stats import mod_str


"""Program to take character definitions and build a PDF of the
character sheet."""

bold_re = re.compile(r'\*\*([^*]+)\*\*')
it_re = re.compile(r'\*([^*]+)\*')
tt_re = re.compile(r'``([^`]+)``')
# A dice string, with optinal backticks: ``1d6 + 3``
dice_re = re.compile(r'`*(\d+d\d+(?:\s*\+\s*\d+)?)`*')

def rst_to_latex(rst):
    """Basic markup of RST to LaTeX code."""
    if rst is None:
        tex = ""
    else:
        tex = rst
        for c in ['\\']:
            tex = tex.replace(c, '\\' + c)
        tex = bold_re.sub(r'\\textbf{\1}', tex)
        tex = it_re.sub(r'\\textit{\1}', tex)
        tex = dice_re.sub(r'\\texttt{\1}', tex)
        tex = tt_re.sub(r'\\texttt{\1}', tex)
        for c in ['#', '$', '%', '&', '~', '_', '^']:
            tex = tex.replace(c, '\\' + c)
    return tex


jinja_env = Environment(
    loader=PackageLoader('dungeonsheets', 'forms'),
    block_start_string='[%',
    block_end_string='%]',
    variable_start_string='[[',
    variable_end_string=']]',
)
jinja_env.filters['rst_to_latex'] = rst_to_latex
jinja_env.filters['mod_str'] = mod_str


CHECKBOX_ON = 'Yes'
CHECKBOX_OFF = 'Off'
PDFTK_CMD = 'pdftk'


def text_box(string):
    """Format a string for displaying in a text box."""
    # remove multiple whitespace without removing linebreaks
    new_string = ' '.join(string.replace('\n', '\m').split())
    # Remove *single* line breaks, swap *multi* line breaks to single (fdf: \r)
    new_string = new_string.replace('\m \m', '\r').replace('\m\m', '\r').replace('\m', ' ')
    return new_string


def create_druid_shapes_pdf(character, basename):
    template = jinja_env.get_template('druid_shapes_template.tex')
    return create_latex_pdf(character, basename, template)


def create_spellbook_pdf(character, basename):
    template = jinja_env.get_template('spellbook_template.tex')
    return create_latex_pdf(character, basename, template)


def create_features_pdf(character, basename):
    template = jinja_env.get_template('features_template.tex')
    return create_latex_pdf(character, basename, template)


def create_latex_pdf(character, basename, template):
    tex = template.render(character=character)
    # Create tex document
    tex_file = f'{basename}.tex'
    with open(tex_file, mode='w') as f:
        f.write(tex)
        
    # Convenience function for removing temporary files
    def remove_temp_files(basename_):
        filenames = [f'{basename_}.tex', f'{basename_}.aux',
                     f'{basename_}.log']
        for filename in filenames:
            if os.path.exists(filename):
                os.remove(filename)
    # Compile the PDF
    pdf_file = f'{basename}.pdf'
    output_dir = os.path.abspath(os.path.dirname(pdf_file))
    try:
        result = subprocess.run(['pdflatex', '--output-directory',
                                 output_dir, tex_file, '-halt-on-error'],
                                stdout=subprocess.DEVNULL, timeout=30)
    except FileNotFoundError:
        # Remove temporary files
        remove_temp_files(basename)
        raise exceptions.LatexNotFoundError()
    else:
        if result.returncode == 0:
            remove_temp_files(basename)
        else:
            raise exceptions.LatexError(f'Processing of {basename}.tex failed.')


def create_spells_pdf(character, basename, flatten=False):
    classes_and_levels = ' / '.join([c.name + ' ' + str(c.level)
                                     for c in character.spellcasting_classes])
    abilities = ' / '.join([c.spellcasting_ability.upper()[:3]
                            for c in character.spellcasting_classes])
    DCs = ' / '.join([str(character.spell_save_dc(c))
                      for c in character.spellcasting_classes])
    bonuses = ' / '.join([mod_str(character.spell_attack_bonus(c))
                          for c in character.spellcasting_classes])
    spell_level = lambda x : (x or 0)
    fields = {
        'Spellcasting Class 2': classes_and_levels,
        'SpellcastingAbility 2': abilities,
        'SpellSaveDC  2': DCs,
        'SpellAtkBonus 2': bonuses,
        # Number of spell slots
        'SlotsTotal 19': spell_level(character.spell_slots(1)),
        'SlotsTotal 20': spell_level(character.spell_slots(2)),
        'SlotsTotal 21': spell_level(character.spell_slots(3)),
        'SlotsTotal 22': spell_level(character.spell_slots(4)),
        'SlotsTotal 23': spell_level(character.spell_slots(5)),
        'SlotsTotal 24': spell_level(character.spell_slots(6)),
        'SlotsTotal 25': spell_level(character.spell_slots(7)),
        'SlotsTotal 26': spell_level(character.spell_slots(8)),
        'SlotsTotal 27': spell_level(character.spell_slots(9)),
    }
    # Cantrips
    cantrip_fields = (f'Spells 10{i}' for i in (14, 16, 17, 18, 19, 20, 21, 22))
    cantrips = (spl for spl in character.spells if spl.level == 0)
    for spell, field_name in zip(cantrips, cantrip_fields):
        fields[field_name] = str(spell)
    # Spells for each level
    field_numbers = {
        1: (1015, 1023, 1024, 1025, 1026, 1027, 1028, 1029, 1030, 1031, 1032, 1033, ),
        2: (1046, 1034, 1035, 1036, 1037, 1038, 1039, 1040, 1041, 1042, 1043, 1044, 1045, ),
        3: (1048, 1047, 1049, 1050, 1051, 1052, 1053, 1054, 1055, 1056, 1057, 1058, 1059, ),
        4: (1061, 1060, 1062, 1063, 1064, 1065, 1066, 1067, 1068, 1069, 1070, 1071, 1072, ),
        5: (1074, 1073, 1075, 1076, 1077, 1078, 1079, 1080, 1081, ),
        6: (1083, 1082, 1084, 1085, 1086, 1087, 1088, 1089, 1090, ),
        7: (1092, 1091, 1093, 1094, 1095, 1096, 1097, 1098, 1099, ),
        8: (10101, 10100, 10102, 10103, 10104, 10105, 10106, ),
        9: (10108, 10107, 10109, 101010, 101011, 101012, 101013),
    }
    prep_numbers = {
        1: (251, 309, 3010, 3011, 3012, 3013, 3014, 3015, 3016, 3017, 3018, 3019, ),
        2: (313, 310, 3020, 3021, 3022, 3023, 3024, 3025, 3026, 3027, 3028, 3029, 3030, ),
        3: (315, 314, 3031, 3032, 3033, 3034, 3035, 3036, 3037, 3038, 3039, 3040, 3041, ),
        4: (317, 316, 3042, 3043, 3044, 3045, 3046, 3047, 3048, 3049, 3050, 3051, 3052, ),
        5: (319, 318, 3053, 3054, 3055, 3056, 3057, 3058, 3059, ),
        6: (321, 320, 3060, 3061, 3062, 3063, 3064, 3065, 3066, ),
        7: (323, 322, 3067, 3068, 3069, 3070, 3071, 3072, 3073, ),
        8: (325, 324, 3074, 3075, 3076, 3077, 3078, ),
        9: (327, 326, 3079, 3080, 3081, 3082, 3083, ),
    }
    for level in field_numbers.keys():
        spells = tuple(spl for spl in character.spells if spl.level == level)
        field_names = tuple(f'Spells {i}' for i in field_numbers[level])
        prep_names = tuple(f'Check Box {i}' for i in prep_numbers[level])
        for spell, field, chk_field in zip(spells, field_names, prep_names):
            fields[field] = str(spell)
            is_prepared = any([spell == Spl
                               for Spl in character.spells_prepared])
            fields[chk_field] = CHECKBOX_ON if is_prepared else CHECKBOX_OFF
        # # Uncomment to post field names instead:
        # for field in field_names:
        #     fields.append((field, field))
    # Make the actual pdf
    dirname = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'forms/')
    src_pdf = os.path.join(dirname, 'blank-spell-sheet-default.pdf')
    make_pdf(fields, src_pdf=src_pdf, basename=basename, flatten=flatten)


def create_character_pdf(character, basename, flatten=False):
    # Prepare the list of fields
    fields = {
        # Character description
        'CharacterName': character.name,
        'ClassLevel': character.classes_and_levels,
        'Background': str(character.background),
        'PlayerName': character.player_name,
        'Race ': str(character.race),
        'Alignment': character.alignment,
        'XP': str(character.xp),
        'Inspiration': str(character.inspiration),
        # Abilities
        'ProfBonus': mod_str(character.proficiency_bonus),
        'STRmod': str(character.strength.value),
        'STR': mod_str(character.strength.modifier),
        'DEXmod ': str(character.dexterity.value),
        'DEX': mod_str(character.dexterity.modifier),
        'CONmod': str(character.constitution.value),
        'CON': mod_str(character.constitution.modifier),
        'INTmod': str(character.intelligence.value),
        'INT': mod_str(character.intelligence.modifier),
        'WISmod': str(character.wisdom.value),
        'WIS': mod_str(character.wisdom.modifier),
        'CHamod': str(character.charisma.value),
        'CHA': mod_str(character.charisma.modifier),
        'AC': str(character.armor_class),
        'Initiative': str(character.initiative),
        'Speed': str(character.speed),
        'Passive': 10 + character.perception,
        # Saving throws (proficiencies handled later)
        'ST Strength': mod_str(character.strength.saving_throw),
        'ST Dexterity': mod_str(character.dexterity.saving_throw),
        'ST Constitution': mod_str(character.constitution.saving_throw),
        'ST Intelligence': mod_str(character.intelligence.saving_throw),
        'ST Wisdom': mod_str(character.wisdom.saving_throw),
        'ST Charisma': mod_str(character.charisma.saving_throw),
        # Skills (proficiencies handled below)
        'Acrobatics': mod_str(character.acrobatics),
        'Animal': mod_str(character.animal_handling),
        'Arcana': mod_str(character.arcana),
        'Athletics': mod_str(character.athletics),
        'Deception ': mod_str(character.deception),
        'History ': mod_str(character.history),
        'Insight': mod_str(character.insight),
        'Intimidation': mod_str(character.intimidation),
        'Investigation ': mod_str(character.investigation),
        'Medicine': mod_str(character.medicine),
        'Nature': mod_str(character.nature),
        'Perception ': mod_str(character.perception),
        'Performance': mod_str(character.performance),
        'Persuasion': mod_str(character.persuasion),
        'Religion': mod_str(character.religion),
        'SleightofHand': mod_str(character.sleight_of_hand),
        'Stealth ': mod_str(character.stealth),
        'Survival': mod_str(character.survival),
        # Hit points
        'HDTotal': character.hit_dice,
        'HPMax': str(character.hp_max),
        # Personality traits and other features
        'PersonalityTraits ': text_box(character.personality_traits),
        'Ideals': text_box(character.ideals),
        'Bonds': text_box(character.bonds),
        'Flaws': text_box(character.flaws),
        'Features and Traits': text_box(character.features_text + character.features_and_traits),
        # Inventory
        'CP': character.cp,
        'SP': character.sp,
        'EP': character.ep,
        'GP': character.gp,
        'PP': character.pp,
        'Equipment': text_box(character.magic_items_text + character.equipment),
    }
    # Check boxes for proficiencies
    ST_boxes = {
        'strength': 'Check Box 11',
        'dexterity': 'Check Box 18',
        'constitution': 'Check Box 19',
        'intelligence': 'Check Box 20',
        'wisdom': 'Check Box 21',
        'charisma': 'Check Box 22',
    }
    for ability in character.saving_throw_proficiencies:
        fields[ST_boxes[ability]] = CHECKBOX_ON
    # Add skill proficiencies
    skill_boxes = {
        'acrobatics': 'Check Box 23',
        'animal_handling': 'Check Box 24',
        'arcana': 'Check Box 25',
        'athletics': 'Check Box 26',
        'deception': 'Check Box 27',
        'history': 'Check Box 28',
        'insight': 'Check Box 29',
        'intimidation': 'Check Box 30',
        'investigation': 'Check Box 31',
        'medicine': 'Check Box 32',
        'nature': 'Check Box 33',
        'perception': 'Check Box 34',
        'performance': 'Check Box 35',
        'persuasion': 'Check Box 36',
        'religion': 'Check Box 37',
        'sleight_of_hand': 'Check Box 38',
        'stealth': 'Check Box 39',
        'survival': 'Check Box 40',
    }
    for skill in character.skill_proficiencies:
        try:
            fields[skill_boxes[skill.replace(' ', '_').lower()]] = CHECKBOX_ON
        except KeyError:
            raise KeyError(f"Unknown skill: '{skill}'")
    # Add weapons
    weapon_fields = [('Wpn Name', 'Wpn1 AtkBonus', 'Wpn1 Damage'),
                     ('Wpn Name 2', 'Wpn2 AtkBonus ', 'Wpn2 Damage '),
                     ('Wpn Name 3', 'Wpn3 AtkBonus  ', 'Wpn3 Damage '),]
    if len(character.weapons) == 0:
        character.wield_weapon('unarmed')
    for _fields, weapon in zip(weapon_fields, character.weapons):
        name_field, atk_field, dmg_field = _fields
        fields[name_field] = weapon.name
        fields[atk_field] = '{:+d}'.format(weapon.attack_modifier)
        fields[dmg_field] = f'{weapon.damage}/{weapon.damage_type}'
    # Other attack information
    attack_str = f'Armor: {character.armor}'
    attack_str += '\n \n'
    attack_str += f'Shield: {character.shield}'
    attack_str += '\n \n'
    attack_str += character.attacks_and_spellcasting
    fields['AttacksSpellcasting'] = text_box(attack_str)
    # Other proficiencies and languages
    prof_text = "Proficiencies:\n" + text_box(character.proficiencies_text)
    prof_text += "\n\nLanguages:\n" + text_box(character.languages)
    fields['ProficienciesLang'] = prof_text
    # Prepare the actual PDF
    dirname = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'forms/')
    src_pdf = os.path.join(dirname, 'blank-character-sheet-default.pdf')
    return make_pdf(fields, src_pdf=src_pdf, basename=basename,
                    flatten=flatten)


def make_pdf(fields: dict, src_pdf: str, basename: str, flatten: bool=False):
    """Create a new PDF by applying fields to a src PDF document.
    
    Parameters
    ==========
    fields :
      Data to fill into the form. The keys are field names in the PDF
      form, and the values will be entered as the value in the PDF.
    src_pdf :
      Path to the PDF that will serve as the template.
    basename :
      The path of the destination PDF without the file extensions. The
      resulting pdf will be {basename}.pdf
    flatten :
      If truthy, the PDF will be collapsed so it is no longer
      editable.
    
    """
    try:
        result = _make_pdf_pdftk(fields, src_pdf, basename, flatten)
    except FileNotFoundError:
        # pdftk could not run, so alert the user and use pdfrw
        warnings.warn(f'Could not run `{PDFTK_CMD}`, using fallback; '
                      'forcing `--editable`.',
                      RuntimeWarning)
        _make_pdf_pdfrw(fields, src_pdf, basename, flatten)


def _make_pdf_pdfrw(fields: dict, src_pdf: str, basename: str, flatten: bool=False):
    """Backup make_pdf function in case pdftk is not available."""
    template = pdfrw.PdfReader(src_pdf)
    # Different types of PDF fields
    BUTTON = '/Btn'
    # Names for entries in PDF annotation list
    DEFAULT_VALUE = '/DV'
    APPEARANCE = '/MK'
    FIELD = '/T'
    PROPS = '/P'
    TYPE = '/FT'
    FLAGS = '/Ff'
    SUBTYPE = '/Subtype'
    ALL_KEYS = ['/DV', '/F', '/FT', '/Ff', '/MK', '/P', '/Rect',
                '/Subtype', '/T', '/Type']
    annots = template.pages[0]['/Annots']
    # Update each annotation if it's in the requested dictionary
    for annot in annots:
        this_field = annot[FIELD][1:-1]
        # Check if the field has a new value passed
        if this_field in fields.keys():
            val = fields[this_field]
            # Convert integers to strings
            if isinstance(val, int):
                val = str(val)
            log.debug(f"Set field '{this_field}' "
                      f"({annot[TYPE]}) "
                      f"to `{val}` ({val.__class__}) "
                      f"in file '{basename}.pdf'")
            # Prepare a PDF dictionary based on the fields properties
            if annot[TYPE] == BUTTON:
                # Radio buttons require special appearance streams
                if val == CHECKBOX_ON:
                    val = bytes(val, 'utf-8')
                    pdf_dict = pdfrw.PdfDict(V=val, AS=val)
                else:
                    continue
            else:
                # All other widget types
                pdf_dict = pdfrw.PdfDict(V=val)
            annot.update(pdf_dict)
        else:
            log.debug(f"Skipping unused field '{this_field}' in file '{basename}.pdf'")
    # Now write the PDF to the new pdf file
    pdfrw.PdfWriter().write(f'{basename}.pdf', template)


def _make_pdf_pdftk(fields, src_pdf, basename, flatten=False):
    """More robust way to make a PDF, but has a hard dependency."""
    # Create the actual FDF file
    fdfname = basename + '.fdf'
    
    fdf = forge_fdf("", fields, [], [], [])
    fdf_file = open(fdfname, "wb")
    fdf_file.write(fdf)
    fdf_file.close()
        # Build the final flattened PDF documents
    dest_pdf = basename + '.pdf'
    popenargs = [
        PDFTK_CMD, src_pdf, 'fill_form', fdfname, 'output', dest_pdf,
    ]
    if flatten:
        popenargs.append('flatten')
    subprocess.call(popenargs)
    # Clean up temporary files
    os.remove(fdfname)
    
    
def make_sheet(character_file, character=None, flatten=False):
    """Prepare a PDF character sheet from the given character file.
    
    Parameters
    ----------
    character_file : str
        File (.py) to load character from. Will save PDF using same name
    character : Character, optional
        If provided, will not load from the character file, just use file
        for PDF name
    flatten : bool, optional
        If true, the resulting PDF will look better and won't be fillable form.
    """
    if character is None:
        character = _char.Character.load(character_file)
    # Set the fields in the FDF
    char_base = os.path.splitext(character_file)[0] + '_char'
    sheets = [char_base + '.pdf']
    pages = []
    char_pdf = create_character_pdf(character=character, basename=char_base,
                                    flatten=flatten)
    pages.append(char_pdf)
    if character.is_spellcaster:
        # Create spell sheet
        spell_base = '{:s}_spells'.format(
            os.path.splitext(character_file)[0])
        create_spells_pdf(character=character, basename=spell_base, flatten=flatten)
        sheets.append(spell_base + '.pdf')
    if len(character.features) > 0:
        feat_base = '{:s}_feats'.format(
            os.path.splitext(character_file)[0])
        try:
            create_features_pdf(character=character, basename=feat_base)
        except exceptions.LatexNotFoundError as e:
            log.warning('``pdflatex`` not available. Skipping features book '
                        f'for {character.name}')
        else:
            sheets.append(feat_base + '.pdf')
    if character.is_spellcaster:
        # Create spell book
        spellbook_base = os.path.splitext(character_file)[0] + '_spellbook'
        try:
            create_spellbook_pdf(character=character, basename=spellbook_base)
        except exceptions.LatexNotFoundError as e:
            log.warning('``pdflatex`` not available. Skipping spellbook '
                        f'for {character.name}')
        else:
            sheets.append(spellbook_base + '.pdf')
    # Create a list of Druid wild_shapes
    wild_shapes = getattr(character, 'wild_shapes', [])
    if len(wild_shapes) > 0:
        shapes_base = os.path.splitext(character_file)[0] + '_wild_shapes'
        try:
            create_druid_shapes_pdf(character=character, basename=shapes_base)
        except exceptions.LatexNotFoundError as e:
            log.warning('``pdflatex`` not available. Skipping wild shapes list '
                        f'for {character.name}')
        else:
            sheets.append(shapes_base + '.pdf')
    # Combine sheets into final pdf
    final_pdf = os.path.splitext(character_file)[0] + '.pdf'
    merge_pdfs(sheets, final_pdf, clean_up=True)


def merge_pdfs(src_filenames, dest_filename, clean_up=False):
    """Merge several PDF files into a single final file.
    
    src_filenames
      Iterable of source PDF file paths to use.
    dest_filename
      Path to requested PDF filename, will be overwritten if it
      exists.
    clean_up : optional
      If truthy, the ``src_filenames`` will be deleted once the
      ``dest_filename`` has been created.
    
    """
    popenargs = (PDFTK_CMD, *src_filenames, 'cat', 'output', dest_filename)
    try:
        subprocess.call(popenargs)
    except FileNotFoundError:
        warnings.warn(f'Could not run `{PDFTK_CMD}`; skipping file concatenation.',
                      RuntimeWarning)
    else:
        # Remove temporary files
        if clean_up:
            for sheet in src_filenames:
                os.remove(sheet)


load_character_file = _char.read_character_file


def main():
    # Prepare an argument parser
    parser = argparse.ArgumentParser(
        description='Prepare Dungeons and Dragons character sheets as PDFs')
    parser.add_argument('filename', type=str, nargs="?",
                        help="Python file with character definition")
    parser.add_argument('--editable', '-e', action="store_true",
                        help="Keep the PDF fields in place once processed.")
    parser.add_argument('--debug', '-d', action="store_true",
                        help="Provide verbose logging for debugging purposes.")
    args = parser.parse_args()
    # Prepare logging if necessary
    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    # Process the requested files
    if args.filename is None:
        filenames = [f for f in os.listdir('.') if os.path.splitext(f)[1] == '.py']
    else:
        filenames = [args.filename]
    for filename in filenames:
        print(f"Processing {os.path.splitext(filename)[0]}...", end='')
        try:
            make_sheet(character_file=filename, flatten=(not args.editable))
        except exceptions.CharacterFileFormatError as e:
            # Only raise the failed exception if this file is explicitly given
            print('invalid')
            if args.filename:
                raise
        except Exception as e:
            print('failed')
            raise
        else:
            print("done")


if __name__ == '__main__':
    main()



from .features import Feature
from math import ceil


# PHB
class RogueExpertise(Feature):
    """At 1st level, choose two of your skill proficiencies, or one of your skill
    proficiencies and your proficiency with thieves’ tools. Your proficiency
    bonus is doubled for any ability check you make that uses either of the
    chosen proficiencies. 

    At 6th level, you can choose two more of your
    proficiencies (in skills or with thieves’ tools) to gain this benefit.

    Add these skills to "skill_expertise" in your character.py file

    """
    name = "Expertise"
    source = "Rogue"


class SneakAttack(Feature):
    """Beginning at 1st level, you know how to strike subtly and exploit a foe’s
    distraction. Once per turn, you can deal an extra 1d6 damage to one
    creature you hit with an attack if you have advantage on the attack
    roll. The attack must use a finesse or a ranged weapon.

    You don’t need advantage on the attack roll if another enemy of the target
    is within 5 feet of it, that enemy isn’t incapacitated, and you don’t have
    disadvantage on the attack roll.

    The amount of the extra damage increases as you gain levels in this class,
    as shown in the Sneak Attack column of the Rogue table.
    """
    _name = "Sneak Attack"
    source = "Rogue"

    @property
    def name(self):
        level = self.owner.Rogue.level
        dice = ceil(level / 2.)
        name = self._name + " ({:d}d6)".format(dice)
        return name


class CunningAction(Feature):
    """Starting at 2nd level, your quick thinking and agility allow you to move
    and act quickly. You can take a bonus action on each of your turns in
    combat. This action can be used only to take the Dash, Disengage, or Hide
    action.

    """
    name = "Cunning Action"
    source = "Rogue"


class UncannyDodge(Feature):
    """Starting at 5th level, when an attacker that you can see hits you with an
    attack, you can use your reaction to halve the attack’s damage against you.

    """
    name = "Uncanny Dodge"
    source = "Class (many)"


class Evasion(Feature):
    """Beginning at 7th level, you can nimbly dodge out o f the way of certain
    area effects, such as a red dragon’s fiery breath or an ice storm
    spell. When you are subjected to an effect that allows you to make a
    Dexterity saving throw to take only half damage, you instead take no damage
    if you succeed on the saving throw, and only half damage if you fail.

    """
    name = "Evasion"
    source = "Class (many)"


class ReliableTalent(Feature):
    """By 11th level, you have refined your chosen skills until they approach
    perfection. Whenever you make an ability check that lets you add your
    proficiency bonus, you can treat a d20 roll of 9 or lower as a 10.

    """
    name = "Reliable Talent"
    source = "Rogue"


class BlindSense(Feature):
    """Starting at 14th level, if you are able to hear, you are aware of the
    location of any hidden or invisible creature within 10 feet of you.
    """
    name = "Blind Sense"
    source = "Rogue"


class SlipperyMind(Feature):
    """By 15th level, you have acquired greater mental strength. You gain
    proficiency in W isdom saving throws.

    """
    name = 'Slippery Mind'
    source = "Rogue"


class Elusive(Feature):
    """Beginning at 18th level, you are so evasive that attackers rarely gain the
    upper hand against you. No attack roll has advantage against you while you
    aren’t incapacitated.

    """
    name = "Elusive"
    source = "Rogue"


class StrokeOfLuck(Feature):
    """At 20th level, you have an uncanny knack for succeeding when you need
    to. If your attack m isses a target within range, you can turn the miss
    into a hit. Alternatively, if you fail an ability check, you can treat the
    d20 roll as a 20.

    Once you use this feature, you can’t use it again until you finish a short
    or long rest.

    """
    name = "Stroke of Luck"
    source = "Rogue"


# Thief
class FastHands(Feature):
    """Starting at 3rd level, you can use the bonus action granted by your Cunning
    Action to make a Dexterity (Sleight of Hand) check, use your thieves’ tools
    to disarm a trap or open a lock, or take the Use an Object action.

    """
    name = "Fast Hands"
    source = "Rogue (Thief)"

    
class SecondStoryWork(Feature):
    """When you choose this archetype at 3rd level, you gain the ability to climb
    faster than normal; climbing no longer costs you extra movement. In
    addition, when you make a running jump, the distance you cover increases by
    a number of feet equal to your Dexterity modifier.

    """
    name = "Second-Story Work"
    source = "Rogue (Thief)"


class SupremeSneak(Feature):
    """Starting at 9th level, you have advantage on a Dexterity (Stealth) check if
    you move no more than half your speed on the same turn

    """
    name = "Supreme Sneak"
    source = "Rogue (Thief)"


class UseMagicDevice(Feature):
    """By 13th level, you have learned enough about the workings of magic that you
    can improvise the use of items even when they are not intended for you. You
    ignore all class, race, and level requirements on the use of magic items

    """
    name = "Use Magic Device"
    source = "Rogue (Thief)"


class ThiefsReflexes(Feature):
    """When you reach 17th level, you have become adept at laying ambushes and
    quickly escaping danger. You can take two turns during the first round of
    any combat. You take your first turn at your normal initiative and your
    second turn at your initiative minus 10. You can’t use this feature when
    you are surprised.

    """
    name = "Thief's Reflexes"
    source = "Rogue (Thief)"


# Assassin
class Assassinate(Feature):
    """Starting at 3rd level, you are at your deadliest when you get the drop on
    your enemies. You have advantage on attack rolls against any creature that
    hasn’t taken a turn in the combat yet. In addition, any hit you score
    against a creature that is surprised is a critical hit.

    """
    name = "Assassinate"
    source = "Rogue (Assassin)"


class InfiltrationExpertise(Feature):
    """Starting at 9th level, you can unfailingly create false identities for
    yourself. You must spend seven days and 25 gp to establish the history,
    profession, and affiliations for an identity. You can’t establish an
    identity that belongs to someone else. For example, you might acquire
    appropriate clothing, letters of introduction, and official- looking
    certification to establish yourself as a member of a trading house from a
    remote city so you can insinuate yourself into the company of other wealthy
    merchants. Thereafter, if you adopt the new identity as a disguise, other
    creatures believe you to be that person until given an obvious reason not
    to.

    """
    name = "Infiltration Expertise"
    source = "Rogue (Assassin)"


class Imposter(Feature):
    """At 13th level, you gain the ability to unerringly mimic another person’s
    speech, writing, and behavior. You must spend at least three hours studying
    these three components of the person’s behavior, listening to speech,
    examining handwriting, and observing mannerisms. Your ruse is indiscernible
    to the casual observer. If a wary creature suspects something is amiss, you
    have advantage on any Charisma (Deception) check you make to avoid
    detection

    """
    name = "Imposter"
    source = "Rogue (Assassin)"


class DeathStrike(Feature):
    """Starting at 17th level, you become a master of instant death. When you
    attack and hit a creature that is surprised, it must make a Constitution
    saving throw (DC 8 + your Dexterity modifier + your proficiency bonus). On
    a failed save, double the damage of your attack against the creature

    """
    name = "Death Strike"
    source = "Rogue (Assassin)"


# Arcane Trickster
class MageHandLegerdemain(Feature):
    """Starting at 3rd level, when you cast mage hand, you can make the spectral
    hand invisible, and you can perform the following additional tasks with it:
    
    -- You can stow one object the hand is holding in a container worn or
    carried by another creature.
    
    -- You can retrieve an object in a container worn or carried by another
    creature.

    --You can use thieves’ tools to pick locks and disarm traps at range

    You can perform one of these tasks without being noticed by a creature if
    you succeed on a Dexterity (Sleight of Hand) check contested by the
    creature’s Wisdom (Perception) check. In addition, you can use the bonus
    action granted by your Cunning Action to control the hand.

    """
    name = "Mage Hand Legerdemain"
    source = "Rogue (Arcane Trickster)"


class MagicalAmbush(Feature):
    """Starting at 9th level, if you are hidden from a creature when you cast a
    spell on it, the creature has disadvantage on any saving throw it makes
    against the spell this turn

    """
    name = "Magical Ambush"
    source = "Rogue (Arcane Trickster)"


class VersatileTrickster(Feature):
    """At 13th level, you gain the ability to distract targets with your mage
    hand. As a bonus action on your turn, you can designate a creature within 5
    feet of the spectral hand created by the spell. Doing so gives you
    advantage on attack rolls against that creature until the end of the turn.

    """
    name = "Versatile Trickster"
    source = "Rogue (Arcane Trickster)"


class SpellThief(Feature):
    """At 17th level, you gain the ability to magically steal the knowledge of how
    to cast a spell from another spellcaster. Immediately after a creature
    casts a spell that targets you or includes you in its area of effect, you
    can use your reaction to force the creature to make a saving throw with its
    spellcasting ability modifier. The DC equals your spell save DC.

    On a failed save, you negate the spell’s effect against you, and you steal
    the knowledge of the spell if it is at least 1st level and of a level you
    can cast (it doesn’t need to be a wizard spell). For the next 8 hours, you
    know the spell and can cast it using your spell slots. The creature can’t
    cast that spell until the 8 hours have passed. Once you use this feature,
    you can’t use it again until you finish a long rest

    """
    name = "Spell Thief"
    source = "Rogue (Arcane Trickster)"


# Inquisitive
class EarForDeceit(Feature):
    """When you choose this archetype at 3rd level, you de— velop a talent for
    picking out lies. Whenever you make a Wisdom (Insight) check to determine
    whether a creature is lying, treat a roll of 7 or lower on the c120 as an
    8.

    """
    name = "Ear for Deceit"
    source = "Rogue (Inquisitive)"


class EyeForDetail(Feature):
    """Starting at 3rd level, you can use a bonus action to make a Wisdom
    (Perception) check to spot a hidden creature or object or to make an
    Intelligence (Investigation) check to uncover or decipher clues

    """
    name = "Eye for Detail"
    source = "Rogue (Inquisitive)"


class InsightfulFighting(Feature):
    """At 3rd level, you gain the ability to decipher an opponent's tactics and
    develop a counter to them. As a bonus action, you can make a Wisdom
    (Insight) check against a creature you can see that isn’t incapacitated,
    contested by the target’s Charisma (Deception) check. If you suc— ceed, you
    can use your Sneak Attack against that target even ifyou don’t have
    advantage on the attack roll, but not if you have disadvantage on it. This
    benefit lasts for 1 minute or until you successfully use this feature
    against a different target

    """
    name = "Insightful Fighting"
    source = "Rogue (Inquisitive)"


class SteadyEye(Feature):
    """Starting at 9th level, you have advantage on any Wisdom (Perception) or
    Intelligence (Investigation) check if you move no more than half your speed
    on the same turn

    """
    name = "Steady Eye"
    source = "Rogue (Inquisitive)"

    
class UnerringEye(Feature):
    """Beginning at 13th level, your senses are almost im« possible to foil. As an
    action, you sense the presence of illusions, shapechangers not in their
    original form, and other magic designed to deceive the senses within 30
    feet ofyou, provided you aren’t blinded or deafened. You sense that an
    effect is attempting to trick you, but you gain no insight into what is
    hidden or into its true nature. You can use this feature a number of times
    equal to your Wisdom modifier (minimum of once), and you regain all
    expended uses of it when you finish a long rest

    """
    name = "Unerring Eye"
    source = "Rogue (Inquisitive)"


class EyeForWeakness(Feature):
    """At 17th level, you learn to exploit a creature’s weak— nesses by carefully
    studying its tactics and movement. While your Insightful Fighting feature
    applies to a creature, your Sneak Attack damage against that creature
    increases by 3d6

    """
    name = "Eye for Weakness"
    source = "Rogue (Inquisitive)"


# Mastermind
class MasterOfIntrigue(Feature):
    """When you choose this archetype at 3rd level, you gain proficiency with the
    disguise kit, the forgery kit, and one gaming set Of your choice. You also
    learn two languages of your choice. Additionally, you can unerringly mimic
    the speech patterns and accent of a creature that you hear speak for at
    least 1 minute, enabling you to pass yourself off as a native speaker of a
    particular land, provided that you know the language.

    """
    name = "Master of Intrigue"
    source = "Rogue (Mastermind)"


class MasterOfTactics(Feature):
    """Starting at 3rd level, you can use the Help action as a bonus
    action. Additionally, when you use the Help action to aid an ally in
    attacking a creature, the target of that attack can be within 30 feet of
    you, rather than within 5 feet of you, if the target can see or hear you

    """
    name = "Master of Tactics"
    source = "Rogue (Mastermind)"

    
class InsightfulManipulator(Feature):
    """Starting at 9th level, if you spend at least 1 minute observing or
    interacting with another creature outside combat, you can learn certain
    information about its ca— pabilities compared to your own. The DM tells you
    if the creature is your equal, superior, or inferior in regard to two of
    the following characteristics of your choice:

    -- Intelligence score

    -- Wisdom score

    -- Charisma score

    --Class levels (if any)

    At the DM’s option, you might also realize you know a piece of the
    creature’s history or one of its personality traits, if it has any

    """
    name = "Insightful Manipulator"
    source = "Rogue (Mastermind)"


class Misdirection(Feature):
    """Beginning at 13th level, you can sometimes cause another creature to
    suffer an attack meant for you. When you are targeted by an attack while a
    creature within 5 feet of you is granting you cover against that attack,
    you can use your reaction to have the attack target that crea— ture instead
    of you

    """
    name = "Misdirection"
    source = "Rogue (Mastermind)"


class SoulOfDeceit(Feature):
    """Starting at 17th level, your thoughts can’t be read by telepathy or other
    means, unless you allow it. You can present false thoughts by succeeding on
    a Charisma (Deception) check contested by the mind reader’s Wis— dom
    (Insight) check. Additionally, no matter what you say, magic that would
    determine if you are telling the truth indicates you are being truthful if
    you so choose, and you can’t be compelled to tell the truth by magic

    """
    name = "Soul of Deceit"
    source = "Rogue (Masterind)"


# Scout
class Skirmisher(Feature):
    """Starting at 3rd level, you are difficult to pin down during a fight. You
    can move up to halfyour speed as a reaction when an enemy ends its turn
    within 5 feet of you. This movement doesn’t provoke opportunity attacks

    """
    name = "Skirmisher"
    source = "Rogue (Scout)"


class Survivalist(Feature):
    """When you choose this archetype at 3rd level, you gain proficiency in the
    Nature and Survival skills if you don’t already have it. Your proficiency
    bonus is doubled for any ability check you make that uses either of those
    pro— ficiencies

    """
    
    def __init__(self, owner=None):
        super().__init__(owner=owner)
        self.owner.skill_expertise += ("nature", "survival")


class SuperiorMobility(Feature):
    """At 9th level, your walking speed increases by 10 feet. If you have a
    climbing or swimming speed, this increase applies to that speed as well.

    """
    name = "Superior Mobility"
    source = "Rogue (Scout)"
    needs_implementation = True  # apply to climbing and swimming


class AmbushMaster(Feature):
    """Starting at 13th level, you excel at leading ambushes and acting first in a
    fight. You have advantage on initiative rolls. In addition, the first
    creature you hit during the first round of a combat becomes easier for you
    and others to strike; attack rolls against that target have advantage until
    the start ofyour next turn

    """
    name = "Ambush Master"
    source = "Rogue (Scout)"
        

class SuddenStrike(Feature):
    """Starting at 17th level, you can strike with deadly speed. If you take the
    Attack action on your turn, you can make one additional attack as a bonus
    action. This attack can benefit from your Sneak Attack even if you have
    already used it this turn, but you can’t use your Sneak Attack against the
    same target more than once in a turn

    """
    name = "Sudden Strike"
    source = "Rogue (Scout)"


# Swashbuckler
class FancyFootwork(Feature):
    """When you choose this archetype at 3rd level, you learn how to land a strike
    and then slip away without reprisal. During your turn, if you make a melee
    attack against a creature, that creature can’t make opportunity attacks
    against you for the rest ofyour turn

    """
    name = "Fancy Footwork"
    source = "Rogue (Swashbuckler)"


class RakishAudacity(Feature):
    """Starting at 3rd level, your confidence propels you into battle. You can give
    yourself a bonus to your initiative rolls equal to your Charisma
    modifier. You also gain an additional way to use your Sneak Attack; you
    don’t need advantage on the attack roll to use your Sneak Attack against a
    creature if you are within 5 feet of it, no other creatures are within 5
    feet of you, and you don’t have disadvantage on the attack roll. All the
    other rules for Sneak Attack still apply to you.

    """
    name = "Rakish Audacity"
    source = "Rogue (Swashbuckler)"


class Panache(Feature):
    """At 9th level, your charm becomes extraordinarily be— guiling. As an action,
    you can make a Charisma (Persuasion) check contested by a creature’s
    Wisdom (Insight) check. The creature must be able to hear you, and the
    two ofyou must share a language. If you succeed on the check and the
    creature is hostile to you, it has disadvantage on attack rolls against
    targets other than you and can’t make opportunity attacks against targets
    other than you.

    This effect lasts for 1 minute, until one of your companions attacks the
    target or affects it with a spell, or until you and the target are more
    than 60 feet apart. If you succeed on the check and the creature isn’t
    hostile to you, it is charmed by you for 1 minute. While charmed, it
    regards you as a friendly acquaintance. This effect ends immediately if you
    or your companions do anything harmful to it

    """
    name = "Panache"
    source = "Rogue (Swashbuckler)"


class ElegantManeuver(Feature):
    """Starting at 13th level, you can use a bonus action on your turn to gain
    advantage on the next Dexterity (Ac— robatics) or Strength (Athletics)
    check you make during the same turn

    """
    name = "Elegant Maneuver"
    source = "Rogue (Swashbuckler)"


class MasterDuelist(Feature):
    """Beginning at 17th level, your mastery of the blade lets you turn failure
    into success in combat. Ifyou miss with an attack roll, you can roll it
    again with advantage. Once you do so, you can’t use this feature again
    until you finish a short or long rest

    """
    name = "Master Duelist"
    source = "Rogue (Swashbuckler)"
    




from .spells import Spell


class Knock(Spell):
    """Choose an object that you can see within range. The object can be a door, a box,
     a chest, a set of manacles, a padlock, or another object that contains a 
    mundane or magical means that prevents access.
    
    A target that is held shut by a 
    mundane lock or that is stuck or barred becomes unlocked, unstuck, or unbarred. 
    If the object has multiple locks, only one of them is unlocked.
    
    If you choose a
     target that is held shut with arcane lock, that spell is suppressed for 10 
    minutes, during which time the target can be opened and shut normally.
    
    When you
     cast the spell, a loud knock, audible from as far away as 300 feet, emanates 
    from the target object.
    """
    name = "Knock"
    level = 2
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ('V',)
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Bard', 'Sorcerer', 'Wizard')


